# Description: A set of fetch methods to get macroeconomic data

# Dependencies
from fredapi import Fred
import pandas as pd
from datetime import datetime, timedelta
import json
import pickle
# Used for downloading data from the internet
import requests
import zipfile
import tempfile
import shutil

# Get current directory
import sys, os
CWD = os.path.dirname(os.path.dirname(__file__))
sys.path.append(CWD)

# Package imports
import timeseries as mt

# Directory for data cache
CACHE_DIR = os.path.join(os.path.expanduser('~'), '.macro_cache')

def parse_year(year: str) -> datetime:
    '''
    Get year, month, day from %Y.%f
    '''
    year_int = int(year)
    month = int((year % 1) * 12) + 1
    day = int(((year % 1) * 12) % 1 * 30) + 1 
    return datetime(year_int, month, day)

def get_fred(data_sources: dict[str, str], start_date: str = "", end_date: str = "", api_key=None, force_fetch=False, verbose=False):
    '''
    Get data from FRED.

    Parameters
    ----------
    data_sources : dict[str, str] or list[str]
        Dictionary or list of FRED series IDs and names to assign to the data. If the value is None, the series ID is used as the name.
    date_one : str, optional
        Start date for data. Default is "".
    date_two : str, optional
        End date for data. Default is "".
    api_key : str, optional
        API key for FRED. Default is None (not needed if using cached data).
    force_fetch : bool, optional
        Force fetching data from FRED. Default is False.
    verbose : bool, optional
        Print verbose output. Default is False.

    Returns
    -------
    mt.TimeseriesTable
        Table of FRED data.
    
    Notes
    -----
    - Data is automatically set to year/quarter/month-end frequency and reindexed daily.
    - Caching is used to store data for one week before fetching new data from FRED (use force_fetch=True to ignore cache).
    '''
    try:
        # Look for api key 
        if api_key is None: # Load API key
            try:
                with open(f'{os.path.dirname(__file__)}/fred_api_key.txt') as f:
                    api_key = f.read().strip()
            except FileNotFoundError as e:
                try: 
                    api_key = os.environ['FRED_API_KEY']
                except KeyError as e:
                    pass # api_key is None
        
        # Create cache directory
        cache_dir = os.path.join(CACHE_DIR, 'fred_data')
        if not os.path.exists(cache_dir):
            print(f"Cache directory does not exist. Creating cache directory at {cache_dir}")
            os.makedirs(cache_dir)

        # Create connection to fred 
        if api_key:
            fred_connection = Fred(api_key=api_key)
        else:
            fred_connection = None
                
        # Results table
        data = mt.TimeseriesTable()

        # Get data from FRED
        current_series_id = None
        for series_id in data_sources:
            # map series_id to data_name
            if type(data_sources) == list:
                data_name = series_id
            elif type(data_sources) == dict:
                data_name = data_sources[series_id]
                if not data_name:
                    data_name = series_id
            else:
                raise ValueError("data_sources must be a list or dictionary.")
            
            # Log current series
            if verbose:
                print(f'Fetching data for {data_name} ({series_id})')
            current_series_id = series_id

            # Check if data is in cache
            cache_file = os.path.join(cache_dir, f"{series_id}.pkl")
            metadata_file = os.path.join(cache_dir, f"{series_id}_metadata.json")
            if not force_fetch and os.path.exists(cache_file) and os.path.exists(metadata_file):
                with open(metadata_file, 'r') as f:
                    metadata = json.load(f)
                    # Validation logic
                    cache_expiry = datetime.strptime(metadata['expiry_date'], '%Y-%m-%d')
                    if datetime.now() < cache_expiry:
                        # Read data from cache
                        with open(cache_file, 'rb') as f:
                            cached_data = pickle.load(f)
                        data[data_name] = cached_data
                        if verbose:
                            print(f"Data for {data_name} loaded from cache.\n")
                        continue
            
            # Get data from FRED (cache is expired or invalid)
            if fred_connection:
                raw_data = fred_connection.get_series(series_id)
                data_freq = fred_connection.get_series_info(series_id)['frequency']
            # Get data from FRED web page
            else: 
                # Add deprecation warning
                print("DeprecationWarning: Fetching data from FRED without an API key is deprecated. Please provide an API key to avoid this warning.", DeprecationWarning)
                page_url = f'https://fred.stlouisfed.org/series/{series_id}/downloaddata/{series_id}.csv'
                try:
                    raw_data = pd.read_csv(page_url, index_col='DATE', parse_dates=True)
                    # cast to pd.Series
                    raw_data = pd.Series(raw_data.squeeze()).dropna()
                except Exception as e:
                    raise ValueError(f"Error fetching data from FRED: {e}")
                # Try to infer frequency
                try:
                    data_freq =  pd.infer_freq(raw_data.index)
                except Exception as e:
                    raise ValueError(f"Error inferring frequency for {series_id}: {e}")
            
            # Save start and end dates
            cache_start_date = raw_data.index[0]
            cache_end_date = raw_data.index[-1]
            # Create timeseries
            new_data = mt.Timeseries(raw_data, name=data_name, source_freq=data_freq, data_source="FRED")
            # new_data.transformations.append('reindex_end_period')
            # Set date range
            if not start_date:
                start_date = new_data.index[0]
            if not end_date:
                end_date = new_data.index[-1]
            new_data = new_data.truncate(start_date, end_date)
            # Save data to table
            data[data_name] = new_data
            if verbose:
                print(f"Data for {data_name} fetched from FRED.\n")

            # Save data to cache
            with open(cache_file, 'wb') as f:
                pickle.dump(new_data, f)
            metadata = {
                'created_date': datetime.now().strftime('%Y-%m-%d'),
                'expiry_date': (datetime.now() + timedelta(weeks=1)).strftime('%Y-%m-%d'),
                'start_date': cache_start_date.strftime('%Y-%m-%d'),
                'end_date': cache_end_date.strftime('%Y-%m-%d')
            }
            with open(metadata_file, 'w') as f:
                json.dump(metadata, f)

        return data
    
    except ValueError as e:
        if '404' in str(e):
            raise ValueError(f"Unable to fetch {series_id}. Check the series ID or try again with API key.")
        else:
            raise e
        
    except Exception as e:
        if force_fetch == False: # try again in case cache is old
            print(f"Error fetching {current_series_id} from FRED. Trying again with force_fetch=True.")
            return get_fred(data_sources, start_date, end_date, api_key, force_fetch=True)
        else:
            raise e

def get_barnichon(filepath: str, table: mt.TimeseriesTable, input_var: str, output_name: str):
    '''
    Get vacancy rate data from Barnichon.

    Parameters
    ----------
    filepath : str
        Path to Barnichon file.
    table : mt.TimeseriesTable
        Table to store data.
    input_var : str
        Input variable to use.
    output_name : str
        Name to assign to the output variable.

    Returns
    -------
    mt.TimeseriesTable
        Table with vacancy rate data.

    Notes
    -----
    - Data is automatically set to quarterly frequency.
    - Necessary data is automatically fetched from FRED if needed.
    '''
    # read data from barnichon file
    barnichon_data = pd.read_csv(filepath)
    barnichon_data['year'] = barnichon_data['year'].apply(parse_year)
    barnichon_data.set_index('year', inplace=True)
    barnichon_data = mt.TimeseriesTable(barnichon_data)

    match input_var: # construct data
        case 'V_LF':
            if 'JTSJOL' not in table:
                table['V'] = get_fred(['JTSJOL'], ['V'])['V']
            if 'CLF16OV' not in table:
                table['L'] = get_fred(['CLF16OV'], ['L'])['L']
            # Create vacancy rate with JOLTS data
            table[output_name] = mt.Timeseries(100 * table['V'] / table['L'], name=output_name, source_freq='Q', data_source='Barnichon')
            # Replace pre-2001 with Barnichon's vacancy rate
            input_data = barnichon_data.df.loc[:'2000-12-31'][input_var]
            for idx in input_data.index:
                if idx <= datetime(2000, 12, 31):
                    table[output_name][idx] = input_data[idx]
        case 'V_hwi':
            table[output_name] = mt.Timeseries(barnichon_data[input_var], name=output_name, source_freq='Q', data_source='Barnichon')
        case _:
            raise ValueError(f'Unknown input_var: {input_var}')
        
    return table

def get_ludvigson():
    '''
    Get macroeconomics uncertaintity data from Ludvigson.

    Returns
    -------
    dict
        Dictionary of TimeseriesTables for each data source (FinancialUncertainty, MacroUncertainty, RealUncertainty).

    Notes
    -----
    - Data is automatically set to monthly frequency.
    - Caching is used to store data for one week.
    '''
    # Define data source
    DATA_SOURCE = 'https://www.sydneyludvigson.com/s/MacroFinanceUncertainty_202408Update.zip'
    
    # Find the cache 
    cache_dir = os.path.join(CACHE_DIR, 'ludvigson_data')
    if not os.path.exists(cache_dir):
        print(f"Cache directory does not exist. Creating cache directory at {cache_dir}")
        os.makedirs(cache_dir)

    # Check if data is in cache
    data_files = ['FinancialUncertainty', 'MacroUncertainty', 'RealUncertainty']
    data = {}
    for data_file in data_files:
        cache_file = os.path.join(cache_dir, f"{data_file}.pkl")
        metadata_file = os.path.join(cache_dir, f"{data_file}_metadata.json")
        if os.path.exists(cache_file) and os.path.exists(metadata_file):
            with open(metadata_file, 'r') as f:
                metadata = json.load(f)
                cache_expiry = datetime.strptime(metadata['expiry_date'], '%Y-%m-%d')
                # Make sure the cache is still valid
                if datetime.now() < cache_expiry:
                    with open(cache_file, 'rb') as f:
                        cached_data = pickle.load(f)
                        data[data_file] = cached_data
                    print(f"Data for {data_file} loaded from cache.")
                    continue
    
    # Check if all data is in cache
    if len(data) == len(data_files):
        return data
    
    # Clear old files from the cache
    shutil.rmtree(cache_dir)
    os.makedirs(cache_dir)
    # Download data from Ludvigson
    response = requests.get(DATA_SOURCE)
    zip_path = os.path.join(tempfile.gettempdir(), 'temp.zip')
    # Save the zip file to a temporary file
    with open(zip_path, 'wb') as f:
        f.write(response.content)
    # Extract the contents of the zip file
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(cache_dir)

    # Parse the contents of the zip file
    file_names = os.listdir(cache_dir)
    for source_file in file_names:
        # Get the data
        data_name = [elem for elem in data_files if elem in source_file][0]
        data_path = os.path.join(cache_dir, source_file)
        src_data = pd.read_excel(data_path)
        # Construct TimeseriesTable
        table = mt.TimeseriesTable()
        # Set index
        src_data.set_index('Date', inplace=True)
        # Create timeseries
        for col in src_data.columns:
            table[col] = mt.Timeseries(src_data[col], name=col, source_freq='monthly', data_source='Ludvigson')
        data[data_name] = table
        # Save data to cache
        cache_file = os.path.join(cache_dir, f"{data_name}.pkl")
        with open(cache_file, 'wb') as f:
            pickle.dump(table, f)
        metadata = {
            'created_date': datetime.now().strftime('%Y-%m-%d'),
            'expiry_date': (datetime.now() + timedelta(weeks=1)).strftime('%Y-%m-%d')
        }
        metadata_file = os.path.join(cache_dir, f"{data_name}_metadata.json")
        with open(metadata_file, 'w') as f:
            json.dump(metadata, f)
        print(f"Data for {data_name} fetched from Ludvigson.")
    
    return data



# import libraries
import unittest
import pandas as pd
import random

# get Timeseries class
import os, sys
script_dir = os.path.dirname(__file__)
sys.path.append(os.path.join(script_dir, '..', 'src', 'macroecon_tools'))
from timeseries import Timeseries, TimeseriesTable

# get fetch data
import fetch_data as fd

class TestTimeseries(unittest.TestCase):
    def setUp(self):
        # Sample data for testing
        self.sample_data = pd.Series([i for i in range(24)], index=pd.date_range("2020-01-01", periods=24, freq="ME"))
        self.ts = Timeseries(self.sample_data, name="Sample", freq="monthly")

    def test_copy(self):
        ts_copy = self.ts.copy()
        # Assert that the copied instance is not the same as the original
        self.assertIsNot(ts_copy, self.ts)
        # Assert that the data and attributes of the copied instance are equal to the original
        self.assertTrue(ts_copy.equals(self.ts))
        self.assertEqual(ts_copy.name, self.ts.name)
        self.assertEqual(ts_copy.get_freqstr(), self.ts.get_freqstr())
        self.assertEqual(ts_copy.transformations, self.ts.transformations)

def test_print():
    sample_data = pd.Series([1, 2, 3, 4], index=pd.date_range("2020-01-01", periods=4, freq="ME"))
    ts = Timeseries(sample_data, name="Sample", source_freq="monthly")
    print(ts)
    large_sample_data = pd.Series([i for i in range(24)], index=pd.date_range("2020-01-01", periods=24, freq="ME"))
    ts = Timeseries(large_sample_data, name="Sample", source_freq="monthly")
    print(ts)

def test_agg():
    index = pd.date_range(start="2020-01-01", periods=36, freq="ME")
    data = Timeseries(pd.Series([i for i in range(36)], index=index), name="test")
    # print(data)
    data = data.aggregate('yearly', 'mean')
    print(f"{data}")

def test_pd_agg():
    index = pd.date_range(start="2020-01-01", periods=36, freq="ME")
    data = pd.Series([i for i in range(36)], index=index)
    data = data.resample('Y').mean()
    print(data)

def test_weekly():
    index = pd.date_range(start="2020-01-01", periods=36, freq="W")
    data = pd.Series([i for i in range(36)], index=index)
    print(Timeseries(data, name='test').aggregate('weekly', 'mean'))

def test_getdata():
    # Test get_fred
    data_sources = ['GDPC1', 'CPIAUCSL']
    data_names = ['GDP', 'CPI']
    data = fd.get_fred(data_sources, data_names)
    print(data)
    data.save(f'{script_dir}/test_data')

def test_load_data():
    # Load data
    data = TimeseriesTable().load(f'{script_dir}/test_data')
    print(f"{data}")

def test_dropna():
    index = pd.date_range(start="2020-01-01", periods=36, freq="ME")
    data = Timeseries(pd.Series([i for i in range(36)], index=index), name='test')
    data = data.dropna()
    print(data)

def test_table():
    index = pd.date_range(start="2020-01-01", periods=36, freq="ME")
    data = Timeseries(pd.Series([i for i in range(36)], index=index), name='test').aggregate('monthly', 'mean')
    table = TimeseriesTable()
    table['test'] = data
    print(table)

def test_pane_multi():
    PERIODS=12
    idx_one = pd.date_range(start="2010-01-01", periods=PERIODS, freq="ME")
    data_one = Timeseries(pd.Series([i for i in range(PERIODS)], index=idx_one), name='test_one').aggregate('monthly', 'mean')
    idx_two = pd.date_range(start="2010-01-01", periods=PERIODS, freq="ME")
    data_two = Timeseries(pd.Series([i*i for i in range(PERIODS)], index=idx_two), name='test_two').aggregate('monthly', 'mean')
    table = TimeseriesTable({
        'test_one': data_one,
        'test_two': data_two
    })
    idx_three = pd.date_range(start="2010-01-01", periods=PERIODS, freq="W-SUN")
    for idx in range(3):
        table[f"test_{idx}"] = Timeseries(pd.Series([i*i*i for i in range(PERIODS)], index=idx_three), name=f'test_{idx}').aggregate('weekly', 'mean')
    print(f"first table\n {table}\n")
    print(f"{table.corr()}")
    print(f"{table.to_latex()}")

def test_ts():
    data = pd.Series([1 + 0.3 * i for i in range(10)], index=pd.date_range('2024-01-01', periods=10))
    ts = Timeseries(data, name="Sample Data", source_freq="daily", data_source="Generated")

    # Apply transformations
    print(ts)
    log_diff = ts.logdiff(1)
    print(log_diff)
    truncated = ts.truncate('2024-01-01', '2024-01-04')
    print(truncated)

def test_corr():
    # generate panel
    PERIODS = 10
    panel = TimeseriesTable()
    for i in range(5):
        data = pd.Series([random.random() for i in range(PERIODS)], index=pd.date_range('2024-01-01', periods=PERIODS))
        ts = Timeseries(data, name=f"Sample Data {i}", source_freq="daily", data_source="Generated")
        panel[f"Data_{i}"] = ts
    # print(panel)
    # print()
    # print(panel.corr())
    # print()
    # print(panel.corr('Data_2', 'Data_1'))
    # print()
    # print(panel.corr('Data_2'))
    # print(panel.to_markdown())
    print(panel.corr("Data_1", "Data_4"))
    print(panel.corr("Data_4", "Data_1"))

def test_moments():
    # generate panel
    PERIODS = 10
    panel = TimeseriesTable()
    for i in range(5):
        data = pd.Series([i for i in range(PERIODS)], index=pd.date_range('2024-01-01', periods=PERIODS))
        ts = Timeseries(data, name=f"Sample Data {i}", source_freq="daily", data_source="Generated")
        panel[f"Data_{i}"] = ts
    print(panel.data_moments())

def test_tst_trunc():
    # Generate data
    PERIODS = 10
    panel = TimeseriesTable()
    for i in range(5):
        data = pd.Series([i for i in range(PERIODS)], index=pd.date_range('2024-01-01', periods=PERIODS, freq='ME'))
        ts = Timeseries(data, name=f"Sample Data {i}", source_freq="daily", data_source="Generated")
        panel[f"Data_{i}"] = ts
    # truncate
    print(panel.truncate('2024-01-01', '2024-01-04'))
    # reindex quarterly
    print(panel.reindex('Q', 'ffill'))

def test_tst_agg():
    # Generate data
    PERIODS = 10
    panel = TimeseriesTable()
    for i in range(5):
        data = pd.Series([i for i in range(PERIODS)], index=pd.date_range('2024-01-01', periods=PERIODS, freq='ME'))
        ts = Timeseries(data, name=f"Sample Data {i}", source_freq="daily", data_source="Generated")
        panel[f"Data_{i}"] = ts
    # aggregate
    print(panel.aggregate('quarterly', 'mean'))

def test_shift():
    # Generate data
    PERIODS = 10
    test_series = Timeseries(pd.Series([i for i in range(PERIODS)], index=pd.date_range('2024-01-01', periods=PERIODS, freq='ME')), name='Test Series')
    # shift
    print(test_series)
    print(test_series.diff(1))

def test_print_tst():
    # generate timeseriestable
    PERIODS = 100
    table = TimeseriesTable()
    for i in range(5):
        data = pd.Series([i for i in range(PERIODS)], index=pd.date_range('2024-01-01', periods=PERIODS))
        ts = Timeseries(data, name=f"Sample Data {i}", source_freq="daily", data_source="Generated")
        table[f"Data_{i}"] = ts
    print(table)

if __name__ == '__main__':
    # test_agg()
    # test_pd_agg()
    # test_print()
    # unittest.main()
    # test_load_data()
    # test_load_data()
    # test_weekly()
    # test_dropna()
    # test_getdata()
    # test_table()
    # test_pane_multi()
    # test_ts()
    # test_corr()
    # test_moments()
    # test_tst_trunc()
    # test_tst_agg()
    # test_shift()
    test_print_tst()


import matplotlib.pyplot as plt
from timeseries import TimeseriesTable

class TimeseriesVisualizer:
    """
    A class for visualizing timeseries data using various plotting methods.

    Attributes
    ----------
    data : TimeseriesTable
        The timeseries data to be visualized.
    """

    def __init__(self, data: TimeseriesTable):
        """
        Initialize the TimeseriesVisualizer with a TimeseriesTable.

        Parameters
        ----------
        data : TimeseriesTable
            The timeseries data to be visualized.
        """
        self.data = data

    def subplotpad(self, top=0.1, left=0.1, vspace=0.1, hspace=0.1, right=0.05, bottom=0.1):
        """
        Adjust subplot positions.

        Parameters:
            top: Top margin (default: 0.11)
            left: Left margin (default: 0.075)
            vspace: Vertical space between subplots (default: 0.065)
            hspace: Horizontal space between subplots (default: 0.07)
            right: Right margin (default: 0.02)
            bottom: Bottom margin (default: 0.04)
        """
        # Get all axes (subplots) in the current figure, sorted top-left to bottom-right
        axes = sorted(plt.gcf().get_axes(), key=lambda ax: (-ax.get_position().y0, ax.get_position().x0))

        # Number of columns, and rows
        ncols = len({ax.get_position().x0 for ax in axes})
        nrows = len({ax.get_position().y0 for ax in axes})

        # Calculate width and height of each subplot
        width = (1 - left - hspace * (ncols - 1) - right) / ncols
        height = (1 - top - vspace * (nrows - 1) - bottom) / nrows

        # Adjust positions of each subplot
        for isub, ax in enumerate(axes):
            irow, icol = divmod(isub, ncols)  # Row and column indices
            left_pos = left + icol * (width + hspace)
            bottom_pos = 1 - top - (irow + 1) * height - irow * vspace
            ax.set_position([left_pos, bottom_pos, width, height])

    def subplots(self, save_path: str, variables: list[str] = None, start_date: str = "", end_date: str = "", is_percent: bool = False):
        """
        Creates a plot of multiple timeseries data in separate subplots.

        Parameters
        ----------
        save_path : str
            The path to save the plot.
        variables : list[str], optional
            A list of variables to plot. If None, all variables in the data are plotted.
        start_date : str, optional
            The start date for the plot.
        end_date : str, optional
            The end date for the plot.
        is_percent : bool, optional
            Whether to format the y-axis as percentages.

        Notes
        -----
        - Plot titles and axis labels can be adjusted with varialbe.set_label() and variable.set_percent().
        """
        # Use all variables if none are provided
        if variables is None:
            variables = list(self.data.df.columns)
        
        # Default to the full range of the data if no start or end dates are provided
        if not start_date:
            start_date = self.data.df.index[0]
        if not end_date:
            end_date = self.data.df.index[-1]

        # Determine the number of columns based on the number of variables
        num_vars = len(variables)
        if num_vars > 6: # three columns
            num_cols = 3
        elif num_vars >= 3: # two columns
            num_cols = 2
        else: # single column
            num_cols = 1
        num_rows = (num_vars + num_cols - 1) // num_cols  # Calculate the number of rows needed

        # Create subplots for each variable
        fig, axs = plt.subplots(num_rows, num_cols, figsize=(6.5, 4))

        # Make the axes iterable even if there is only one subplot
        if num_vars > 1:
            axs = axs.flatten()
        else:
            axs = [axs]

        for idx, ax in enumerate(axs):
            # Stop plotting if all variables have been plotted
            if idx >= num_vars:
                ax.set_visible(False)
                continue
            
            # Plot each variable in its respective subplot
            ax.plot(self.data[variables[idx]][start_date:end_date])
            # Set the title using the variable label
            ax.set_title(self.data[variables[idx]].label)
            # Format the y-axis based on the is_percent flag
            if self.data[variables[idx]].is_percent or is_percent:
                ax.yaxis.set_major_formatter('{x:.0f}%')
            else:
                ax.yaxis.set_major_formatter('{x:.0f}')
            # Add grid and adjust subplot layout
            ax.grid()
            ax.autoscale(tight=True)
            ax.margins(y=0.05)  # Add padding to y-axis
            # Remove years from x-axis for inner plots
            if idx < num_vars - num_cols:
                ax.set_xticklabels([])

        # Optimize layout and save the plot
        plt.tight_layout()
        self.subplotpad()
        plt.savefig(save_path)
        plt.close()

    def two_vars(self, save_path: str, x_var: str, y_var: str, title: str = "", start_date: str = "", end_date: str = "", x_is_percent: bool = False, y_is_percent: bool = False):
        """
        Plot two variables against each other in one graph.

        Parameters
        ----------
        save_path : str
            The path to save the plot.
        x_var : str
            The variable for the x-axis.
        y_var : str
            The variable for the y-axis.
        title : str, optional
            The title of the plot.
        start_date : str, optional
            The start date for the plot.
        end_date : str, optional
            The end date for the plot.
        x_is_percent : bool, optional
            Whether to format the x-axis as percentages.
        y_is_percent : bool, optional
            Whether to format the y-axis as percentages.
        
        Notes
        -----
        - Plot titles and axis labels can be adjusted with varialbe.set_label() and variable.set_percent().
        """
        # Determine the overlap of the two variables in the specified date range
        if not start_date:
            start_date = max(self.data.df[x_var].index[0], self.data.df[y_var].index[0])
        if not end_date:
            end_date = min(self.data.df[x_var].index[-1], self.data.df[y_var].index[-1])

        # Create the plot
        fig, ax = plt.subplots(figsize=(6.5, 4))
        ax.scatter(self.data.df[x_var][start_date:end_date], self.data.df[y_var][start_date:end_date])

        # Set title and labels
        ax.set_title(title if title else f"{x_var} vs {y_var}")
        ax.set_xlabel(x_var)
        ax.set_ylabel(y_var)
        # Format axes based on percentage flags
        if self.data[x_var].is_percent or x_is_percent:
            ax.xaxis.set_major_formatter('{x:.0f}%')
        else:
            ax.xaxis.set_major_formatter('{x:.0f}')
        if self.data[y_var].is_percent or y_is_percent:
            ax.yaxis.set_major_formatter('{x:.0f}%')
        else:
            ax.yaxis.set_major_formatter('{x:.0f}')
        # Add grid and adjust layout
        ax.grid()
        ax.autoscale(tight=True)
        ax.margins(x=0.05, y=0.05) # Add padding to x and y axes
        plt.savefig(save_path)
        plt.close()

    def multi_lines(self, save_path: str, variables: list[str], title: str, start_date: str = "", end_date: str = "", is_percent: bool = False):
        """
        Plot multiple variables on one graph.

        Parameters
        ----------
        save_path : str
            The path to save the plot.
        variables : list[str], optional
            A list of variables to plot.
            Default: [], plot all variables.
        title : str
            The title of the plot.
        start_date : str, optional
            The start date for the plot.
        end_date : str, optional
            The end date for the plot.
        is_percent : bool, optional
            Whether to format the y-axis as percentages.
        """
        # Default to the full range of the data if no start or end dates are provided
        if not start_date:
            start_date = min(self.data.df[var].index[0] for var in variables)
        if not end_date:
            end_date = max(self.data.df[var].index[-1] for var in variables)
        
        # Check for variables
        if len(variables) == 0:
            for var in self.data:
                variables.append(var)

        # Create the plot
        fig, ax = plt.subplots(figsize=(6.5, 4))

        # Define line styles to cycle through
        line_styles = [('red', '-'), ('blue', '--'), ('darkgreen', '-.'), ('black', ':')]
        # plot data, use default styles if insufficient predefined types
        if len(variables) <= len(line_styles):
            for idx, var in enumerate(variables):
                ax.plot(self.data.df[var][start_date: end_date], color=line_styles[idx][0], linestyle=line_styles[idx][1], label=var)
        else:
            for idx, var in enumerate(variables):
                ax.plot(self.data.df[var][start_date: end_date], label=var)  
   
        # Set title and format y-axis
        ax.set_title(title)
        ax.yaxis.set_major_formatter('{x:.0f}%' if is_percent else '{x:.0f}')
        # Add grid, legend, and adjust layout
        ax.grid()
        ax.autoscale(tight=True)
        ax.legend()
        plt.savefig(save_path)
        plt.close()

    def plot_individual(self, save_path: str, variables: list[str] = [], start_date: str = "", end_date: str = "", is_percent: bool = False):
        """
        Generate individual figures for the variables specified in the data.

        Parameters
        ----------
        save_path : str
            The path to save the plot. Figures saved as f"{save_path}_{variable_name}.png".
        variables : list[str], optional
            A list of variables to plot. 
            Default: [] (plot all variables).
        start_date : str, optional
            The start date for the plot.
            Default: "" (full range).
        end_date : str, optional
            The end date for the plot.
            Default: "" (full range).
        is_percent : bool, optional
            Whether to format the y-axis as percentages.
            Default: False (format as integers).
        """
        # Use all variables if none are provided
        if len(variables) == 0:
            variables = list(self.data.df.columns)

        # Create a plot for each variable
        for var in variables:
             # Default to the full range of the data if no start or end dates are provided
            if not start_date:
                series_start = self.data[var].index[0]
            if not end_date:
                series_end = self.data[var].index[-1]

            # Create the plot
            fig, ax = plt.subplots(figsize=(6.5, 4))
            ax.plot(self.data[var][series_start: series_end])
            ax.set_title(self.data[var].label)
            
            # Format the y-axis based on the is_percent flag
            if self.data[var].is_percent or is_percent:
                ax.yaxis.set_major_formatter('{x:.0f}%')
            else:
                ax.yaxis.set_major_formatter('{x:.0f}')

            # Add grid and adjust layout
            ax.grid()
            ax.autoscale(tight=True)
            ax.margins(y=0.05)

            # Save and close the figure
            plt.savefig(f"{save_path}_{var}.png")
            plt.close()



# Get current directory
# import os, sys
# dir_path = os.path.dirname(os.path.realpath(__file__))
# sys.path.append(dir_path)

# import submodules
from .timeseries import *
from .fetch_data import *
from .visualizer import *



# Description: A class used to represent a timeseries object.

# Data processing
import pandas as pd
import numpy as np
# Timing
from datetime import datetime
from dateutil.relativedelta import relativedelta
# Filters
from quantecon._filter import hamilton_filter as qe_hamilton_filter
from scipy.signal import detrend
from statsmodels.tsa.filters.hp_filter import hpfilter
# Save/Load
import pickle

# Add path to constants
import sys, os
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

# Import constants
from constants import Constants

class Timeseries(pd.Series):
    """
    A class used to represent a timeseries object.

    Attributes
    ----------
    data : pd.Series
        a single variable in a time table
    is_copy : bool
        a flag to indicate if the data is a copy (avoid reindexing to daily)
    name : str
        the name of the variable
    label : str (optional)
        the label of the variable used in visualizations. 
        default is to use the variable name.
    is_percent : bool (optional)
        a flag to indicate if the data is a percent (used in visualizations)
        default is False (not a percent).
    source_freq : str (optional)
        the frequency of the data source of the variable
    data_source : str (optional)
        the source of the data
    transformations : list[str]
        a list of transformations applied to the data

    Methods
    -------
    trans(form, lags=None)
        Transforms the data using the specified form (e.g., 'logdiff', 'diff', 'log', '100log').
        Must provide number of lags for 'logdiff' and 'diff'.
    agg(timestep, method)
        Aggregates the data using the specified method (e.g., 'quarterly', 'monthly', 'yearly').
        Must provide method (e.g., lastvalue, mean, sum).
    filter(method, date_one, date_two, p=None, h=None)
        Filters the data using the specified method (e.g., 'linear' or 'hamilton').
        Must provide start date (date_one) and end date (date_two).
        For 'hamilton', must also provide lag length (p) and lead length (h).
    """
    _metadata = ['name', 'label', 'source_freq', 'data_source', 'is_percent', 'transformations']

    def __init__(self, data, is_copy = False, name: str = None, label: str = "", source_freq: str = "unknown", data_source : str = "unknown", is_percent: bool = False, transformations: list[str] = [], *args, **kwargs):
        """
        Initializes a Timeseries object.

        Parameters
        ----------
        data : pd.Series or array-like
            A single variable in a time table.
        is_copy : bool, optional
            If True, the data is treated as a copy and not reindexed to daily frequency. Default is False.
        name : str, optional
            The name of the variable. Default is None.
        label : str, optional
            The label of the variable used in visualizations. Default is the name.
        source_freq : str, optional
            The frequency of the source data (e.g., 'quarterly', 'monthly', 'yearly'). Default is "unknown".
        data_source : str, optional
            The source of the data. Default is "unknown".
        is_percent: bool, optional
            The time series is a percent (used in visualizations). Default is False (not a percent).
        transformations : list of str, optional
            A list of transformations applied to the data. Default is an empty list.
        *args : 
            Additional positional arguments passed to the pd.Series constructor.
        **kwargs : 
            Additional keyword arguments passed to the pd.Series constructor.

        Raises
        ------
        ValueError
            If the index is not a datetime index.
        ValueError
            If the variable name is not provided.
        ValueError
            If the frequency is not provided.

        Notes
        -----
            - The data is converted to float64 type.
            - If `is_copy` is False, the data is reindexed to daily frequency.
            - The index is converted to a datetime index and its frequency is inferred.
            - Adding a new parameter also needs to be reflected in _update.
        """
        # Make sure all data is valid
        data = data.replace('.', np.nan)
        # convert data to float64
        data = pd.Series(data).dropna().astype('float64')
        # reindex the data to daily if not a copy
        if not is_copy:
            data = data.asfreq('D')
        # Call pd.Series constructor
        super().__init__(data, *args, **kwargs)

        # Metadata
        self.transformations = [val for val in transformations]
        self.source_freq = source_freq
        self.data_source = data_source

        # Ensure the index is a datetime index
        try:
            self.index = pd.to_datetime(self.index)
            # infer the frequency
            try:
                self.index.freq = pd.infer_freq(self.index)
            except Exception as e:
                print(f'Unable to infer frequency.\n{e}')
        except Exception as e:
            raise ValueError(f'Timeseries Class: Index must be a datetime index.\n{e}')
        
        # Check if need to rename
        if name:
            self.name = name
        if not self.name:
            raise ValueError('Timeseries Class: Variable name not provided')
        self.label = label if label else self.name # add the label attribute

        # Add is_percent
        self.is_percent = is_percent

    # copy constructor
    def _update(self, data: pd.Series, transformation: str = None, name: str = ""):
        """
        Updates the current Timeseries object with new data.

        Parameters
        ----------
        data : pd.Series
            The new data to update the Timeseries object with.
        name : str, optional
            The name of the variable. If provided, it is updated.
        transformation : str, optional
            A transformation applied to the new data. If provided, it is added to the list of transformations.

        Returns
        -------
        Timeseries
            A new Timeseries object with the updated data and transformations.
        """
        if transformation:
            self.transformations.append(transformation)
        if name:
            self.name = name
        return Timeseries(data, is_copy=True, name=self.name, label=self.label, source_freq=self.source_freq, data_source=self.data_source, is_percent=self.is_percent, transformations=self.transformations)

    def set_percent(self, is_percent: bool = True):
        """
        Sets the Timeseries object as a percent.

        Parameters
        ----------
        is_percent : bool, optional
            If True, the Timeseries object is set as a percent. Default is True.
        """
        self.is_percent = is_percent
        return self
    
    def set_label(self, label: str):
        """
        Sets the label of the Timeseries object.

        Parameters
        ----------
        label : str
            The label to set for the Timeseries object.
        """
        self.label = label
        return self

    def __finalize__(self, other, method=None, **kwargs):
        """
        Propagate metadata from another Timeseries object to the current instance.

        Parameters
        ----------
        other : Timeseries
            The Timeseries object from which to propagate metadata.
        method : str, optional
            The method used for propagation. Default is None.
        **kwargs : 
            Additional keyword arguments.

        Returns
        -------
        self : Timeseries
            The Timeseries object with propagated metadata.
        """
        if isinstance(other, Timeseries):
            self.name = getattr(other, 'name', None)
            self.source_freq = getattr(other, 'source_freq', None)
            self.data_source = getattr(other, 'data_source', None)
            self.transformations = getattr(other, 'transformations', [])
        return self
    
    # getters
    def get_freqstr(self):
        """
        Returns the frequency of the timeseries.

        Returns
        -------
        str
            The frequency of the timeseries.
        """
        return self.index.freqstr
    
    # override string representation
    def __repr__(self):
        """
        Returns a string representation of the Timeseries object.

        Returns
        -------
        str
            A string representation of the Timeseries object.
        """
        # create data display
        temp_data = self.copy().dropna()
        data = ""
        if len(temp_data) == 0:
            data = "Empty Timeseries\n"
        elif len(temp_data) <= 5:
            for i in range(len(temp_data)):
                data += f"{temp_data.index[i].strftime('%Y-%m-%d')}    {temp_data.iloc[i]}\n"
        else:
            n = 2
            # get the first n rows without the header
            for i in range(n):
                data += f"{temp_data.index[i].strftime('%Y-%m-%d')}    {temp_data.iloc[i]}\n"
            data += f"\t...\n"
            # get the last n rows
            for i in range(len(temp_data) - n, len(temp_data)):
                data += f"{temp_data.index[i].strftime('%Y-%m-%d')}    {temp_data.iloc[i]}\n"

        # build metadata
        metadata = f"Name: {self.name}, Freq: {self.dropna().get_freqstr()}, Length: {len(self.dropna())}"
        if self.source_freq:
            metadata += f", Source Freq: {self.source_freq}"
        if self.data_source:
            metadata += f", Data Source: {self.data_source}"
        metadata += "\n"
        
        # return string
        return (f"{data}"
                f"...\n"
                f"{metadata}"
                f"Transformations: {self.transformations}\n")  

    # save and load
    def save(self, file_path: str):
        '''
        Saves the timeseries to a file.

        Parameters
        ----------
        file_path : str
            The file path to save the timeseries to.
        '''
        # check if file path has pkl extension
        if 'pkl' not in file_path:
            if '.' in file_path:
                file_path = file_path.split('.')[0] + '.pkl'
            else:
                file_path += '.pkl'
        # save the timeseries to a file
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)

    def load(self, file_path):
        '''
        Loads the timeseries from a file.

        Parameters
        ----------
        file_path : str
            The file path to load the timeseries from.
        '''
        # check if file path has pkl extension
        if 'pkl' not in file_path:
            if '.' in file_path:
                file_path = file_path.split('.')[0] + '.pkl'
            else:
                file_path += '.pkl'
        # load the timeseries from a file
        with open(file_path, 'rb') as f:
            self = pickle.load(f) 
        return self

    # Override operators
    def __mul__(self, other):
        """
        Multiplies the data by a scalar.

        Parameters
        ----------
        other : float
            The scalar to multiply the data by.

        Returns
        -------
        Timeseries
            A new Timeseries object with the multiplied data.
        """
        # check if other is a scalar
        if not isinstance(other, (int, float)):
            if other.name is None:
                other_label = f"series_unknown"
            else:
                other_label = f"series_{other.name}"
        else:
            other_label = other
        return self._update(pd.Series(self) * other, f'multiply_{other_label}')

    def __truediv__(self, other):
        """
        Divides the data by a scalar.

        Parameters
        ----------
        other : float
            The scalar to divide the data by.

        Returns
        -------
        Timeseries
            A new Timeseries object with the divided data.
        """
        if not isinstance(other, (int, float)):
            if other.name is None:
                other_label = f"series_unknown"
            else:
                other_label = f"series_{other.name}"
        else:
            other_label = other
        return self._update(pd.Series(self) / other, f'divide_{other_label}')
    
    def __add__(self, other):
        """
        Adds the data to a scalar.

        Parameters
        ----------
        other : float
            The scalar to add to the data.

        Returns
        -------
        Timeseries
            A new Timeseries object with the added data.
        """
        if not isinstance(other, (int, float)):
            if other.name is None:
                other_label = f"series_unknown"
            else:
                other_label = f"series_{other.name}"
        else:
            other_label = other
        return self._update(pd.Series(self) + other, f'add_{other_label}')
    
    def __sub__(self, other):
        """
        Subtracts the data by a scalar.

        Parameters
        ----------
        other : float
            The scalar to subtract from the data.

        Returns
        -------
        Timeseries
            A new Timeseries object with the subtracted data.
        """
        if not isinstance(other, (int, float)):
            if other.name is None:
                other_label = f"series_unknown"
            else:
                other_label = f"series_{other.name}"
        else:
            other_label = other
        return self._update(pd.Series(self) - other, f'subtract_{other_label}')
    
    def parse_date(self, date: str) -> str:
        """
        Accepts multiple string formats and returns a string in the format 'dd-mmm-yyyy'.

        Parameters
        ----------
        date : str
            The date to parse.

        Returns
        -------
        str
            The parsed date in the format 'dd-mmm-yyyy'.
        """
        if isinstance(date, datetime):
            return date.strftime('%d-%b-%Y')
        else:
            if '/' in date: # assume date is in the format 'mm/dd/yyyy'
                print(f"Assuming {date} is formatted as 'mm/dd/yyyy'")
                date = datetime.strptime(date, '%m/%d/%Y').strftime('%d-%b-%Y')
            elif '-' in date and len(date.split('-')[0]) == 4: # assume date is in the format 'yyyy-mm-dd'
                print(f"Assuming {date} is formatted as 'yyyy-mm-dd'")
                date = datetime.strptime(date, '%Y-%m-%d').strftime('%d-%b-%Y')
            elif '-' in date and len(date.split('-')[1]) == 2: # assume date is in the format 'dd-mm-yyyy'
                print(f"Assuming {date} is formatted as 'dd-mm-yyyy'")
                date = datetime.strptime(date, '%d-%m-%Y').strftime('%d-%b-%Y')
            else:
                print(f"Assuming {date} is formatted as 'dd-mmm-yyyy'")
                date = datetime.strptime(date, '%d-%b-%Y').strftime('%d-%b-%Y')
            return date

    # Transform data
    def logdiff(self, nlag: int, freq: str = None):
        """
        Transforms the data using the log difference method.

        Parameters
        ----------
        nlag : int
            The lag length for the transformation.
        freq : str, optional
            Frequency of original data. Default is None.

        Returns
        -------
        Timeseries
            A new Timeseries object with the transformed data.
        """
        trans_freq = self.get_freqstr() if not freq else freq
        annpp = Constants.annscale_map(trans_freq) / nlag
        return self._update(annpp * np.log(self / self.shift(nlag)), f'logdiff_{nlag}_{trans_freq}')
    
    def diff(self, nlag: int):
        """
        Transforms the data using the difference method.

        Parameters
        ----------
        nlag : int
            The lag length for the transformation.

        Returns
        -------
        Timeseries
            A new Timeseries object with the transformed data.
        """
        return self._update(pd.Series(self) - self.shift(nlag), f'diff_{nlag}')
    
    def log(self):
        """
        Transforms the data using the log method.

        Returns
        -------
        Timeseries
            A new Timeseries object with the transformed data.
        """
        return self._update(np.log(self), 'log')
    
    def log100(self):
        """
        Transforms the data using the 100 times log method.

        Returns
        -------
        Timeseries
            A new Timeseries object with the transformed data.
        """
        return self._update(100 * np.log(self), 'log100')
    
    # Aggregation
    def aggregate(self, timestep: str, method: str):
        """
        Aggregates the data using the specified method.

        Parameters
        ----------
        timestep : str
            The timestep to aggregate the data (e.g., 'quarterly', 'monthly', 'yearly').
        method : str
            The aggregation method to use (e.g., 'lastvalue', 'mean', 'sum').

        Returns
        -------
        Timeseries
            A new Timeseries object with the aggregated data.
        """
        # Perform aggregation using super().resample and the specified aggregation method
        aggregated = pd.Series(getattr(super().resample(timestep), Constants.agg_map[method])()).dropna()
        # Update the current Timeseries object with the aggregated data
        if aggregated.index.freq is None:
            aggregated = aggregated.asfreq(timestep)
        # aggregated.index.freq = Constants.freq_map[timestep]
        return self._update(aggregated, f'agg_{timestep}_{method}')

    # Trunctate
    def truncate(self, date_one: str = "", date_two: str = ""):
        """
        Truncates the data between the specified dates.

        Parameters
        ----------
        date_one : str, optional
            The start date in the format 'dd-mmm-yyyy'. Default is "" (start of data).
        date_two : str, optional
            The end date in the format 'dd-mmm-yyyy'. Default is "" (end of data).

        Returns
        -------
        Timeseries
            A new Timeseries object with the truncated data.
        """
        # Check if date_one and date_two are provided
        if not date_one:
            date_one = self.index[0]
        if not date_two:
            date_two = self.index[-1]
        date_one = self.parse_date(date_one)
        date_two = self.parse_date(date_two)
        # update self data
        truncated_data = self[date_one: date_two]
        truncated_data.index = pd.to_datetime(truncated_data.index)
        # return the truncated data
        return self._update(truncated_data, f'trunc_{date_one}_{date_two}')
    
    # Dropna
    def dropna(self):
        """
        Drops missing values from the data.

        Returns
        -------
        Timeseries
            A new Timeseries object with the missing values dropped.
        """
        return self._update(super().dropna())

    # filters
    def linear_filter(self, date_one: str, date_two: str):
        """
        Filters the data using the linear method.

        Parameters
        ----------
        date_one : str
            The start date in the format 'dd-mmm-yyyy'.
        date_two : str
            The end date in the format 'dd-mmm-yyyy'.

        Returns
        -------
        Timeseries
            A new Timeseries object with the filtered data.
        """
        date_one = self.parse_date(date_one)
        date_two = self.parse_date(date_two)

        # Time range
        self = self.truncate(date_one, date_two)
        return self._update(detrend(self, axis=0, type='linear'), f'linear_filter_{date_one}_{date_two}')
    
    def hamilton_filter(self, date_one: str, date_two: str, lag_len: int = None, lead_len: int = None):
        """
        Filters the data using the Hamilton method.

        Parameters
        ----------
        date_one : str
            The start date in the format 'dd-mmm-yyyy'.
        date_two : str
            The end date in the format 'dd-mmm-yyyy'.
        lagLength : int, optional
            The lag length for the 'hamilton' filter. 
            Default is None (1 for yearly, 4 for quarterly, 12 for monthly).
        leadLength : int, optional
            The lead length for the 'hamilton' filter.
            Default is None (2 for yearly, 8 for quarterly, 24 for monthly).

        Returns
        -------
        Timeseries
            A new Timeseries object with the filtered data.

        Raises
        ------
        ValueError
            If the frequency is not supported for the 'hamilton' filter.
        """
        date_one = self.parse_date(date_one)
        date_two = self.parse_date(date_two)

        # get default lag and lead lengths
        if self.get_freqstr() in Constants.year_like:
            lag_len = 1
            lead_len = 2
        elif self.get_freqstr() in Constants.quarter_like:
            lag_len = 4
            lead_len = 8
        elif self.get_freqstr() == Constants.month_like:
            lag_len = 12
            lead_len = 24
        else:
            raise ValueError(f'{self.get_freqstr()} frequency not supported for Hamilton filter')
        
        # get the tstart 
        date_one_time = datetime.strptime(date_one, '%d-%b-%Y')
        if self.get_freqstr() in Constants.year_like:
            tstart = date_one_time - relativedelta(years=(lag_len + lead_len - 1))
        elif self.get_freqstr() in Constants.quarter_like:
            tstart = date_one_time - relativedelta(months=3*(lag_len + lead_len - 1))
        elif self.get_freqstr() in Constants.month_like:
            tstart = date_one_time - relativedelta(months=(lag_len + lead_len - 1))
        trham = self.truncate(tstart, date_two)
        # Get cyclical component 
        cycle, trend = qe_hamilton_filter(trham, lead_len, lag_len)
        cycle_series = pd.Series(cycle.flatten(), index=trham.index).dropna()
        return self._update(cycle_series, f'hamilton_filter_{date_one}_{date_two}_{lag_len}_{lead_len}')

    def hp_filter(self, date_one: str = "", date_two: str = "", lamb: int = None):
        """
        Implements the hp (Hodrick-Prescott) filter.

        Parameters
        ----------
        date_one : str, optional
            The start date in the format 'dd-mmm-yyyy'. 
            Default is "" (start of data).
        date_two : str, optional
            The end date in the format 'dd-mmm-yyyy'.
            Default is "" (end of data).
        lamb : int, optional
            The smoothing parameter for the hp filter.
            Default is None (6.25 for yearly, 1600 for quarterly, 129600  for monthly).
        """
        # Parse dates
        if not date_one:
            date_one = self.index[0]
        if not date_two:
            date_two = self.index[-1]
        date_one = self.parse_date(date_one)
        date_two = self.parse_date(date_two)
        # Time range
        self = self.truncate(date_one, date_two)

        # Get default lambda
        if lamb:
            pass
        elif self.get_freqstr() in Constants.year_like:
            lamb = 6.25
        elif self.get_freqstr() in Constants.quarter_like:
            lamb = 1600
        elif self.get_freqstr() == Constants.month_like:
            lamb = 129600
        else:
            raise ValueError(f'{self.get_freqstr()} frequency not supported for Hamilton filter')
        
        # Apply hp filter
        cycle, trend = hpfilter(self, lamb)
        cycle_series = pd.Series(cycle, index=self.index).dropna()
        return self._update(cycle_series, f'hp_filter_{date_one}_{date_two}_{lamb}')
    
class TimeseriesTable(dict):
    """
    A custom TimeseriesTable that maintains a pd.DataFrame and a dictionary with Timeseries objects.
    """

    def __init__(self, data: dict = {},*args, **kwargs):
        """
        Initializes the TimeseriesTable. Ensures columns are Timeseries instances.

        Parameters
        ----------
        data : dict, optional
            A dictionary where keys are column names and values are Timeseries instances. Default is an empty dictionary.
        """
        super().__init__(data, *args, **kwargs)


    # Update setitem
    def __setitem__(self, key, value):
        """
        Sets the item in the TimeseriesTable.

        Parameters
        ----------
        key : str
            The key to set.
        value : Timeseries
            The Timeseries object to set.
        """
        if not isinstance(value, Timeseries):
            raise ValueError('Value must be a Timeseries object')
        # update name of the Timeseries
        value = value._update(value, name=key)
        # set the item
        super().__setitem__(key, value)

    def __getitem__(self, key):
        data = super().__getitem__(key).dropna()
        return data

    # Build DataFrame
    @property
    def df(self):
        """
        Returns the TimeseriesTable as a DataFrame.

        Returns
        -------
        pd.DataFrame
            A DataFrame representation of the TimeseriesTable.
        """
        # Reindex all columns to daily frequency
        return pd.concat({key: val.asfreq('D') for key, val in self.items()}, axis=1).dropna(how='all')

    # Length
    def __len__(self):
        """
        Returns the number of rows with some data.

        Returns
        -------
        int
            The length of the TimeseriesTable.
        """
        return len(self.df.dropna(how='all'))
    
    # Override attr
    def __getattr__(self, key):
        # if attribute not found try on the DataFrame
        try:
            return getattr(self.df, key)
        except AttributeError:
            raise AttributeError(f"Attribute '{key}' not found in TimeseriesTable")

    def __repr__(self):
        """
        Returns a string representation of the TimeseriesTable object. Displays the first/last two rows.
        """
        # show the first/last two rows
        with pd.option_context('display.max_rows', 4):
            return repr(self.df.dropna(how='all'))
    
    def to_latex(self):
        """
        Returns a LaTeX representation of the TimeseriesTable object.
        """
        return self.df.dropna(how='all').to_latex()
    
    def to_markdown(self):
        """
        Returns a Markdown representation of the TimeseriesTable object.
        """
        return self.df.dropna(how='all').to_markdown()
    
    # Utility
    def apply(self, func, *args, **kwargs):
        """
        Applies a function to the data.

        Parameters
        ----------
        func : function
            The function to apply to the data.

        Returns
        -------
        pd.DataFrame
            A DataFrame with the function applied to the data.
        """
        return TimeseriesTable({key: func(val, *args, **kwargs) for key, val in self.items()})

    # truncate method
    def truncate(self, date_one: str = "", date_two: str = ""):
        """
        Truncates the data between the specified dates.

        Parameters
        ----------
        date_one : str, optional
            The start date in the format 'dd-mmm-yyyy'. Default is "" (start of data).
        date_two : str, optional
            The end date in the format 'dd-mmm-yyyy'. Default is "" (end of data).

        Returns
        -------
        TimeseriesTable
            A new TimeseriesTable object with the truncated data.
        
        Notes
        -----
            - Assumes that date parsing is handled by the Timeseries.truncate method.
        """
        # update self data
        return TimeseriesTable({key: val.truncate(date_one, date_two) for key, val in self.items()})

    # dropna method
    def dropna(self):
        """
        Drops missing values from the data.

        Returns
        -------
        TimeseriesTable
            A new TimeseriesTable object with the missing values dropped.
        """
        return TimeseriesTable({key: val.dropna() for key, val in self.items()})
    
    # reindex method
    def aggregate(self, timestep: str, method: str):
        """
        Aggregates the data using the specified method.

        Parameters
        ----------
        timestep : str
            The timestep to aggregate the data (e.g., 'quarterly', 'monthly', 'yearly').
        method: str
            The aggregation method to use (e.g., 'lastvalue', 'mean', 'sum').
        """
        return TimeseriesTable({key: val.aggregate(timestep, method) for key, val in self.items()})

    # correlation
    def corr(self, var_one: str = "", var_two: str = ""):
        """
        Returns the correlation matrix of the data.

        Parameters
        ----------
        var_one : str, optional
            The first variable to compute the correlation matrix. Default is "".
        var_two : str, optional
            The second variable to compute the correlation matrix. Default is "".

        Returns
        -------
        pd.DataFrame
            A DataFrame with the correlation matrix.

        Notes
        -----
            - If `var_one` and `var_two` are provided, the correlation between the two variables is returned.
            - If only `var_one` is provided, the correlation between `var_one` and all other variables is returned.
        """
        if var_one and var_two:
            return self.df[var_one].corr(self.df[var_two])
        elif var_one:
            return self.df.corrwith(self.df[var_one])
        else:
            return self.df.corr()
        
    # generate requested moments of table
    @staticmethod
    def apply_moment_func(data: Timeseries, moment: str):
        """
        Computes the specified statistical moment for a given Timeseries.

        Parameters
        ----------
        data : Timeseries
            The data to compute the moment on.
        moment : str
            The statistical moment to compute. Options are 'mean', 'SD', 'Skew', 'Kurt', 'min', 'max'.

        Returns
        -------
        pd.Series
            A new Series object with the computed moment.

        Raises
        ------
        ValueError
            If the specified moment is not available.
        """
        if moment == 'mean':
            return data.mean()
            # return lambda X: np.nanmean(X, axis=0)
        elif moment == 'SD':
            return data.std()
            # return lambda X: np.nanstd(X, axis=0, ddof=1)
        elif moment == 'Skew':
            return data.skew()
            # return lambda X: skew(X, axis=0, nan_policy='omit', bias=False)
        elif moment == 'Kurt':
            return data.kurt()
            # return lambda X: kurtosis(X, axis=0, nan_policy='omit', bias=False)
        elif moment == 'min':
            return data.min()
            # return lambda X: np.nanmin(X, axis=0)
        elif moment == 'max':
            return data.max()
            # return lambda X: np.nanmax(X, axis=0)
        else:
            raise ValueError('Moment not available')

    def data_moments(self, vars: list[str] = [], moments: list[str] = []):
        """
        Computes specified statistical moments for given variables in a dataset.

        Parameters
        ----------
        vars : list of str
            The list of variable names for which to display the moments. Default is an empty list (all variables).
        moments : list of str
            The list of moments to compute. Options are 'mean', 'SD', 'Skew', 'Kurt', 'min', 'max'. Default is an empty list (all moments).

        Returns
        -------
        None
            Prints the table of computed moments for the specified variables.
        """
        # Parse input
        if len(vars) == 0:
            vars = self.df.columns
        if len(moments) == 0:
            moments = ['mean', 'SD', 'Skew', 'Kurt', 'min', 'max']

        # Create empty table
        moments_table = pd.DataFrame()

        # Compute moments and append to table
        for moment in moments:
            # apply each fun to the columns 
            vals = []
            for col in vars:
                vals.append(TimeseriesTable.apply_moment_func(self.df[col], moment))
            # append vals as a new row to momtab
            moments_table = pd.concat([moments_table, pd.DataFrame([vals])], ignore_index=True)
        
        # Round and add row labels
        def format_numbers(x):
            return f"{x:.3f}"
        moments_table = moments_table.map(format_numbers)
        moments_table.columns = self.df.columns
        moments_table.index = moments

        # Display table
        return moments_table
    
    # save and load
    def save(self, file_path):
        '''
        Saves the TimeseriesTable to a file.

        Parameters
        ----------
        file_path : str
            The file path to save the TimeseriesTable to.
        '''
        # check if file path has pkl extension
        if 'pkl' not in file_path:
            if '.' in file_path:
                file_path = file_path.split('.')[0] + '.pkl'
            else:
                file_path += '.pkl'
        # save the TimeseriesTable to a file
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)

    def load(self, file_path):
        '''
        Loads the TimeseriesTable from a file.

        Parameters
        ----------
        file_path : str
            The file path to load the TimeseriesTable from.
        '''
        # check if file path has pkl extension
        if 'pkl' not in file_path:
            if '.' in file_path:
                file_path = file_path.split('.')[0] + '.pkl'
            else:
                file_path += '.pkl'
        # load the TimeseriesTable from a file
        with open(file_path, 'rb') as f:
            self = pickle.load(f)
        return self
    
if __name__ == '__main__':
    # List all methods in timeseriestable
    print([func for func in dir(TimeseriesTable) if callable(getattr(TimeseriesTable, func)) and not func.startswith("__")])


class Constants:
    """
    A class to hold constant mappings and scales used throughout the application.

    Attributes
    ----------
    freq_map : dict
        A dictionary mapping time frequencies to their corresponding pandas resample codes.
        Keys are:
            'quarterly' -> 'Q'
            'monthly' -> 'M'
            'yearly' -> 'Y'
    agg_map : dict
        A dictionary mapping aggregation methods to their corresponding pandas aggregation functions.
        Keys are:
            'lastvalue' -> 'last'
            'mean' -> 'mean'
            'sum' -> 'sum'
            'min' -> 'min'
            'max' -> 'max'
    ANNSCALE_MAP : dict
        A dictionary mapping time frequencies to their corresponding annualization scales.
        Keys are:
            'daily' -> 36500
            'weekly' -> 5200
            'monthly' -> 1200
            'quarterly' -> 400
            'yearly' -> 100
            'annual' -> 100
    """

    freq_to_n = ['H', 'D', 'M', 'Q', 'Y']
    agg_map = {
        'lastvalue': 'last',
        'last': 'last',
        'mean': 'mean',
        'average': 'mean',
        'sum': 'sum',
        'min': 'min',
        'max': 'max'
    }
    
    # group similar date formats
    year_like = ['Y', 'YE', 'yearly', 'annual']
    quarter_like = ['Q', 'QE', 'QE-DEC', 'quarterly', 'QS', 'QS-OCT']
    month_like = ['M', 'ME', 'monthly']
    week_like = ['W', 'W-SUN', 'weekly']
    day_like = ['D', 'daily']

    # return annualization scale
    def annscale_map(freq: str):
        """
        A function to return the annualization scale of a given frequency.

        Parameters
        ----------
        freq : str
            The frequency of the data to be annualized.

        Returns
        -------
        int
            The annualization scale of the given frequency.

        Raises
        ------
        ValueError
            If the given frequency is not recognized
        """
        if freq in Constants.year_like:
            return 100
        elif freq in Constants.quarter_like: # quarterly
            return 400
        elif freq in Constants.month_like: # monthly
            return 1200
        elif freq in Constants.week_like:
            return 5200
        elif freq in Constants.day_like:
            return 36500
        else:
            raise ValueError(f"Frequency {freq} not recognized.")

    